/*
	USB mass storage driver for BeOS R5
	Copyright (C) 2000 ITO, Takayuki
	All rights reserved
*/

#include "driver.h"

#define	DPRINTF_INFO(x)	dprintf x
#define	DPRINTF_ERR(x)	dprintf x

static int device_number = 0;

my_device_info *create_device (
	const usb_device *dev, 
	const usb_interface_info *ii,
	const unusual_dev *un_dev)
{
	my_device_info *my_dev = NULL;
	int number;
	area_id area;
	sem_id sem;

	assert (usb != NULL && dev != NULL);

	number = device_number++;

	my_dev = malloc (sizeof (my_device_info));
	if (my_dev == NULL)
		return NULL;

	my_dev->sem_cb = sem = create_sem (0, "usb_mass_cb");
	if (sem < 0)
	{
		DPRINTF_ERR ((MY_ID "create_sem() failed %d\n", (int) sem));
		free (my_dev);
		return NULL;
	}

	my_dev->sem_lock = sem = create_sem (1, "usb_mass_lock");
	if (sem < 0)
	{
		DPRINTF_ERR ((MY_ID "create_sem() failed %d\n", (int) sem));
		delete_sem (my_dev->sem_cb);
		free (my_dev);
		return NULL;
	}

	my_dev->buffer_area = area = create_area ("usb_mass_buffer",
		(void **) &my_dev->buffer, B_ANY_KERNEL_ADDRESS,
		BUF_SIZ, B_CONTIGUOUS, B_READ_AREA | B_WRITE_AREA);
	if (area < 0)
	{
		DPRINTF_ERR ((MY_ID "create_area() failed %d\n", (int) area));
		delete_sem (my_dev->sem_cb);
		delete_sem (my_dev->sem_lock);
		free (my_dev);
		return NULL;
	}

	my_dev->dev = dev;
	my_dev->number = number;
	my_dev->open = 0;
	my_dev->open_fds = NULL;
	my_dev->sector_size = 512;
	if (un_dev != NULL)
	{
		my_dev->subclass = un_dev->subclass;
		my_dev->protocol = un_dev->protocol;
	}
	else
	{
		my_dev->subclass = ii->descr->interface_subclass;
		my_dev->protocol = ii->descr->interface_protocol;
	}
#ifdef DATA_TOGGLE_WORKAROUND
	my_dev->bulkInEptToggle = my_dev->bulkOutEptToggle = 
		my_dev->intrEptToggle = 0;
#endif
	my_dev->active = true;

	switch (my_dev->subclass)
	{
	case USB_SUBCLASS_MASS_SFF8020:
	case USB_SUBCLASS_MASS_SFF8070:
	case USB_SUBCLASS_MASS_UFI:
		my_dev->cdblen = 12;	/* fixed */
		break;
	case USB_SUBCLASS_MASS_RBC:
	case USB_SUBCLASS_MASS_QIC157:
	case USB_SUBCLASS_MASS_SCSI:
	default:
		my_dev->cdblen = 0;		/* command dependent */
		break;
	}

	my_dev->intrEpt = NULL;
	switch (my_dev->protocol)
	{
		case USB_PROTOCOL_MASS_CBI:
		{
			const usb_endpoint_info *ei;
			ei = &ii->endpoint [2];
			/* interrupt IN endpoint */
			assert ((ei->descr->attributes & 3) == 3);
			assert (ei->descr->endpoint_address & 0x80);
			my_dev->intrEpt = ei;
			my_dev->hooks = &cbi_hooks;
			break;
		}

		case USB_PROTOCOL_MASS_CB:
			if (ii->endpoint_count > 2)
				DPRINTF_INFO ((MY_ID "warning: over 2 endpoints despite of CB\n"));
			my_dev->hooks = &cbi_hooks;
			break;

		case USB_PROTOCOL_MASS_BULKONLY:
			if (ii->endpoint_count > 2)
				DPRINTF_INFO ((MY_ID "warning: over 2 endpoints despite of bulk-only\n"));
			my_dev->hooks = &bulkonly_hooks;
			break;
	}

	{
		const usb_endpoint_info *ei0, *ei1;
		ei0 = &ii->endpoint [0];
		ei1 = &ii->endpoint [1];
		/* bulk endpoints */
		assert ((ei0->descr->attributes & 3) == 2);
		assert ((ei1->descr->attributes & 3) == 2);
		/* one is IN and the other is OUT */
		if (ei0->descr->endpoint_address & 0x80)
		{
			assert ((ei1->descr->endpoint_address & 0x80) == 0);
			my_dev->bulkInEpt = ei0;
			my_dev->bulkOutEpt = ei1;
		}
		else
		{
			assert (ei1->descr->endpoint_address & 0x80);
			my_dev->bulkInEpt = ei1;
			my_dev->bulkOutEpt = ei0;
		}
	}

	return my_dev;
}

void remove_device (my_device_info *my_dev)
{
	assert (my_dev != NULL);
	delete_area (my_dev->buffer_area);
	delete_sem (my_dev->sem_cb);
	delete_sem (my_dev->sem_lock);
	free (my_dev);
}
