/*
	USB mass storage driver for BeOS R5
	Copyright (C) 2000 ITO, Takayuki
	All rights reserved
*/

#include "driver.h"

#define	DPRINTF_INFO(x)	dprintf x
#define	DPRINTF_ERR(x)	dprintf x

#define	TIMEOUT_VALUE	(10*1000*1000)	/* microseconds */

static void usb_callback (
	void *cookie, 
	uint32 status, 
	void *data, 
	uint32 actual_len)
{
	my_device_info *my_dev = cookie;
	assert (cookie != NULL);
	my_dev->actual_length = actual_len;
	my_dev->bus_status = status;	/* B_USB_STATUS_* */
	release_sem (my_dev->sem_cb);
}

status_t
do_control (my_device_info *my_dev, int request_type, int request,
	int value, int index, size_t length, void *buf)
{
	status_t st;

	assert (usb != NULL && my_dev != NULL && buf != NULL && 
		length <= BUF_SIZ && my_dev->buffer != NULL && my_dev->dev != NULL);

	DPRINTF_INFO ((MY_ID "do_control()\n"));
	acquire_sem (my_dev->sem_lock);
	if ((request_type & 0x80) == 0)
		memcpy (my_dev->buffer, buf, length);
	my_dev->actual_length = 0;
	my_dev->bus_status = B_USB_STATUS_SUCCESS;
	st = usb->queue_request (my_dev->dev, request_type, request,
		value, index, length, my_dev->buffer, length, usb_callback, my_dev);
	if (st == B_USBD_SUCCESS)
	{
		if (acquire_sem_etc (my_dev->sem_cb, 1, B_RELATIVE_TIMEOUT, 
			(bigtime_t) TIMEOUT_VALUE) == B_TIMED_OUT)
		{
			DPRINTF_ERR ((MY_ID MY_ERR "do_control() timeout\n"));
			/* no way to cancel control xfer explicitly */
		}

		if (request_type & 0x80)
			memcpy (buf, my_dev->buffer, length);
	}
	else
		DPRINTF_ERR ((MY_ID MY_ERR "do_control() failed %d\n", (int) st));

	release_sem (my_dev->sem_lock);
	my_dev->usbd_status = st;
	return (my_dev->usbd_status == B_USBD_SUCCESS && 
		my_dev->bus_status == B_USB_STATUS_SUCCESS) ? B_OK : B_ERROR;
}

status_t do_interrupt (
	my_device_info *my_dev, 
	const usb_endpoint_info *ept, 
	void *buf, size_t length)
{
	status_t st;

	assert (usb != NULL && my_dev != NULL && ept != NULL && buf != NULL && 
		length <= BUF_SIZ && my_dev->dev != NULL && my_dev->buffer != NULL);

	DPRINTF_INFO ((MY_ID "do_interrupt()\n"));
	acquire_sem (my_dev->sem_lock);
/*	if ((ept->descr->endpoint_address & 0x80) == 0)
		memcpy (my_dev->buffer, buf, length); */
	my_dev->actual_length = 0;
	my_dev->bus_status = B_USB_STATUS_SUCCESS;
	st = usb->queue_interrupt (ept->handle, my_dev->buffer, length, 
		usb_callback, my_dev);
	if (st == B_USBD_SUCCESS)
	{
		if (acquire_sem_etc (my_dev->sem_cb, 1, B_RELATIVE_TIMEOUT, 
			(bigtime_t) TIMEOUT_VALUE) == B_TIMED_OUT)
		{
			DPRINTF_ERR ((MY_ID MY_ERR "do_interrupt() timeout\n"));
			/* cancel_queued_transfers() may call callback func
			 w/ B_USB_STATUS_IRP_CANCELLED_BY_REQUEST? */
			usb->cancel_queued_transfers (ept->handle);
			acquire_sem (my_dev->sem_cb);
		}

/*		if (ept->descr->endpoint_address & 0x80) */
			memcpy (buf, my_dev->buffer, length);
	}
	else
		DPRINTF_ERR ((MY_ID MY_ERR "do_interrupt() failed %d\n", (int) st));

	release_sem (my_dev->sem_lock);
	my_dev->usbd_status = st;
	return (my_dev->usbd_status == B_USBD_SUCCESS && 
		my_dev->bus_status == B_USB_STATUS_SUCCESS) ? B_OK : B_ERROR;
}

status_t do_bulk_sg (
	my_device_info *my_dev, 
	const usb_endpoint_info *ept, 
	const iovec *vec, size_t vecsiz, 
	size_t length)
{
	status_t st;
	int i;

	assert (usb != NULL);
	assert (my_dev != NULL);
	assert (ept != NULL);
	assert (vec != NULL);
	assert (my_dev->dev != NULL);
	assert (my_dev->buffer != NULL);
	for (i = 0; i < vecsiz; i++)
		assert (vec [i].iov_base != NULL);

	DPRINTF_INFO ((MY_ID "do_bulk_sg()\n"));
	acquire_sem (my_dev->sem_lock);
	if ((ept->descr->endpoint_address & 0x80) == 0)
	{
		uint8 *dst = my_dev->buffer;
		for (i = 0; i < vecsiz; i++)
		{
			memcpy (dst, vec [i].iov_base, vec [i].iov_len);
			dst += vec [i].iov_len;
		}
	}
	my_dev->actual_length = 0;
	my_dev->bus_status = B_USB_STATUS_SUCCESS;
	st = usb->queue_bulk (ept->handle, my_dev->buffer, length, 
		usb_callback, my_dev);
	if (st == B_USBD_SUCCESS)
	{
		if (acquire_sem_etc (my_dev->sem_cb, 1, B_RELATIVE_TIMEOUT, 
			(bigtime_t) TIMEOUT_VALUE) == B_TIMED_OUT)
		{
			DPRINTF_ERR ((MY_ID MY_ERR "do_bulk() timeout\n"));
			/* cancel_queued_transfers() may call callback func
			 w/ B_USB_STATUS_IRP_CANCELLED_BY_REQUEST? */
			usb->cancel_queued_transfers (ept->handle);
			acquire_sem (my_dev->sem_cb);
		}

		if (ept->descr->endpoint_address & 0x80)
		{
			uint8 *src = my_dev->buffer;
			for (i = 0; i < vecsiz; i++)
			{
				memcpy (vec [i].iov_base, src, vec [i].iov_len);
				src += vec [i].iov_len;
			}
		}
	}
	else
		DPRINTF_ERR ((MY_ID MY_ERR "do_bulk() failed %d\n", (int) st));

	release_sem (my_dev->sem_lock);
	my_dev->usbd_status = st;
	return (my_dev->usbd_status == B_USBD_SUCCESS && 
		my_dev->bus_status == B_USB_STATUS_SUCCESS) ? B_OK : B_ERROR;
}

status_t do_bulk (
	my_device_info *my_dev, 
	const usb_endpoint_info *ept, 
	void *buf, size_t length)
{
	iovec vec;

	vec.iov_base = buf;
	vec.iov_len = length;
	return do_bulk_sg (my_dev, ept, &vec, 1, length);
}

